/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_2.cpp

   Embedding samples
   Adding Default error handler.

   The default error handler will just print the string representation of
   intercepted errors on a provided Falcon Stream. If the user does not
   provide a stream, the error handler will crash; it is necessary to provide
   a minimal setup: Falcon provides a set of stdOutputStream(), stdErrorStream()
   and stdInputStream() object factories that will create a Falcon stream
   pointing to the named system resource.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_2.cpp -o falcon_embed_2

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

// This is the routine that embeds falcon
int embed( const char *script_name )
{
   // Let's create a default error handler
   // It will write on stderr, and the stream will be destroyed with the handler
   Falcon::DefaultErrorHandler err( Falcon::stdErrorStream(), true );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   if ( ! rt.loadFile( script_name ) )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  " or its dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachine vm;
   // we want the default error handler to intercept also VM errors
   vm.errorHandler( &err );

   vm.link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   if( ! vm.link( &rt ) )
   {
      std::cerr << "Sorry, cannot link the script and related modules for " << script_name << std::endl;
      return 1;
   }

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
      std::cerr << "Sorry, cannot launch the script main routine for " << script_name << std::endl;
      return 1;
   }

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   return embed( script_name );
}

/* end of falcon_embed_2.cpp */
