/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_4.cpp

   Embedding samples
   Adding an embedded module

   We show how to compile an embedded module that provides data
   and functions usable from the script.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_4.cpp -o falcon_embed_4

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

//============================================================
// Begin of the embedded module
//============================================================

// some forward decl
FALCON_FUNC say_hi( Falcon::VMachine *vm );

// methods of a class
FALCON_FUNC ProgramInfo_init( Falcon::VMachine *vm );
FALCON_FUNC ProgramInfo_refresh( Falcon::VMachine *vm );

Falcon::Module *createAppModule()
{
   // create a module and give it an unique name
   Falcon::Module *module = new Falcon::Module;
   module->name( "appModule" );

   // now add some constant...
   module->addConstant( "APP_REFRESH_FORWARD", (Falcon::int64) 0 );
   module->addConstant( "APP_REFRESH_BACKWARD", (Falcon::int64) 1 );

   // then add the functions
   module->addExtFunc( "say_hi", say_hi );

   // and add a class:
   // actually, adding the constructor (ProgramInfo_init) is optional
   Falcon::Symbol *cls_pinfo = module->addClass( "ProgramInfo", ProgramInfo_init );

   // add some property
   module->addClassProperty( cls_pinfo, "first" );
   module->addClassProperty( cls_pinfo, "second" );

   // and then add a method
   module->addClassMethod( cls_pinfo, "refresh", ProgramInfo_refresh );

   // the module is complete, we can return it
   return module;
}


FALCON_FUNC say_hi( Falcon::VMachine *vm )
{
   // how many parameters we have?
   Falcon::uint32 params = vm->paramCount();

   // raise an error if we have not enough params
   if ( params == 0 )
   {
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "Please, provide at least one parameter" ) ) );
      return;
   }

   // say hi to all parameters
   for ( Falcon::uint32 p = 0; p < params; p++ )
   {
      // Convert the item; instead of a simple conversion, involve the VM
      Falcon::String itmRep;
      vm->itemToString( itmRep, vm->param( p ) );
      if ( vm->hadError() )
      {
         // we had a conversion error -- and we have raised something
         return;
      }

      vm->stdOut()->writeString( "Say hi to: " );
      vm->stdOut()->writeString( itmRep );
      vm->stdOut()->writeString( "\n" ); // \n is translated into the system line terminator
   }

   // return the number of items done
   vm->retval( (Falcon::int64) params );
}

// How to initialize a simple object
FALCON_FUNC ProgramInfo_init( Falcon::VMachine *vm )
{
   // let's get the parameters and store it in the properties
   Falcon::Item *p1 = vm->param( 0 ); // first param will go in "self.first"
   Falcon::Item *p2 = vm->param( 1 ); // second param will go in "self.second"

   // the self object is readied for us in the VM
   Falcon::CoreObject *self = vm->self().asObject();

   // has the first parameter been provided?
   if ( p1 != 0 )
      self->setProperty( "first", *p1 ); // will be a shallow copy
   else
      self->setProperty( "first", "Uninitialized" );

   if ( p2 != 0 )
      self->setProperty( "second", *p2 ); // will be a shallow copy
   else
      self->setProperty( "second", "Uninitialized" );

   // this is a constructor; there is no need to return anything.
}

// a simple method incrementing or decrementing numeric properties.
FALCON_FUNC ProgramInfo_refresh( Falcon::VMachine *vm )
{
   // this is a method. Get the self item.
   Falcon::CoreObject *self = vm->self().asObject();

   // we want one of our constants as a parameter
   Falcon::Item *refresh_mode = vm->param( 0 ); // first param will go in "self.first"

   if ( refresh_mode == 0 || ! refresh_mode->isOrdinal() )
   {
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "N" ) ) );  // standard falcon code for parameter declaration
      return;
   }

   int mode = (int) refresh_mode->forceInteger();
   if( mode != 0 && mode != 1 )
   {
      // standard falcon code for parameter declaration
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "must be an APP_REFRESH_* constant" ) ) );
      return;
   }

   Falcon::int64 returned = 0;

   Falcon::Item first;
   if ( self->getProperty( "first", first ) && first.isOrdinal() )
   {
      first.setInteger( first.forceInteger() + ( mode == 0 ? 1 : -1 ) );
      self->setProperty( "first", first );
      returned += first.asInteger();
   }

   Falcon::Item second;
   if ( self->getProperty( "second", second ) && second.isOrdinal() )
   {
      second.setInteger( second.forceInteger() + ( mode == 0 ? 1 : -1 ) );
      self->setProperty( "second", second );
      returned += second.asInteger();
   }

   vm->retval( returned );
}

//============================================================
// End of the embedded module
//============================================================

// This is the routine that embeds falcon
int embed( const char *script_name, int argc, char **argv )
{
   // Let's create a default error handler
   // It will write on stderr, and the stream will be destroyed with the handler
   Falcon::DefaultErrorHandler err( Falcon::stdErrorStream(), true );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   if ( ! rt.loadFile( script_name ) )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  " or its dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachine vm;
   // we want the default error handler to intercept also VM errors
   vm.errorHandler( &err );

   vm.link( Falcon::core_module_init() );  // add the core module

   // Finally add our application module
   Falcon::Module *appMod = createAppModule();
   vm.link( appMod );

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   if( ! vm.link( &rt ) )
   {
      std::cerr << "Sorry, cannot link the script and related modules for " << script_name << std::endl;
      return 1;
   }

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm.findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm.findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm.findGlobalItem( "args" );

   // get the topmost (and so, the main) module, just to set the correct name.
   const Falcon::Module *mainMod = vm.mainModule()->module();

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::GarbageString( &vm, mainMod->name() );
   *scriptPath = new Falcon::GarbageString( &vm, script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray( &vm );
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::GarbageString( &vm, argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
      std::cerr << "Sorry, cannot launch the script main routine for " << script_name << std::endl;
      return 1;
   }

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. However, if the VM exited because
   // of runtime errors, the value in A should be ignored.
   if ( ! vm.hadError() )
   {
      Falcon::String str_regA;
      vm.regA().toString( str_regA );

      // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
      Falcon::AutoCString c_regA( str_regA );
      std::cout << "VM Output: " << c_regA.c_str() << std::endl;
   }
   else {
      std::cout << "VM exited with error; we won't show its output. " << std::endl;
   }

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );
   
   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   return embed( script_name, argc - 2, argv + 2 );
}

/* end of falcon_embed_4.cpp */
