/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_cb3.cpp

   Embedding samples
   VM callback support - moveable callback points

   It is useful to provide the scripts with the ability to
   change dynamically their handlers. This may be done by
   reading the entry point after the Main code execution,
   where the script has the chance to change the
   handlers.

   Actually, a script may do that also from an object initializer,
   that is called at link time, so before we have the chance to
   load any caller, but the solution to have the script run its
   main routine before reading the handlers is more elegant.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_cb3.cpp -o falcon_embed_cb3

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

// Forward declaration.
// We're using this function to retreive the callable item directly
// from the module data and the VM data.
Falcon::GarbageLock *getCallable(
      Falcon::VMachine *vm,
      const Falcon::String &callable );

// This is the routine that embeds falcon
int embed( const char *script_name, int argc, char **argv )
{

   // Let's create a default error handler
   // It will write on stderr, and the stream will be destroyed with the handler
   Falcon::DefaultErrorHandler err( Falcon::stdErrorStream(), true );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );
   if ( ! mainMod )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  "." << std::endl;
      return 1;
   }

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   if ( ! rt.addModule( mainMod ) )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  " or its dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachine vm;
   // we want the default error handler to intercept also VM errors
   vm.errorHandler( &err );

   vm.link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   if( ! vm.link( &rt ) )
   {
      std::cerr << "Sorry, cannot link the script and related modules for " << script_name << std::endl;
      return 1;
   }

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm.findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm.findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm.findGlobalItem( "args" );

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::GarbageString( &vm, mainMod->name() );
   *scriptPath = new Falcon::GarbageString( &vm, script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray( &vm );
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::GarbageString( &vm, argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   //======================================
   // find the callbacks in the main module
   //
   Falcon::GarbageLock *cbOnLoad = getCallable( &vm, "onLoad" );

   //===================================
   // Launch the on load function
   //
   if ( cbOnLoad != 0 )
   {
      // we may use iOnLoad as well, but we show here the usage of
      // cbOnLoad as THAT is the object that should be stored safe
      // somewhere in a real program.

      // call without parameters
      vm.callItem( cbOnLoad->item(), 0 );

      // raised an error?
      if ( vm.hadError() )
         return 1;

      // reset the event, that should be a normal "exit on return frame".
      vm.resetEvent();

      // we're not calling cbOnLoad anymore, so this is a good point to
      // allow the VM to dispose it at will
      vm.gcUnlock( cbOnLoad );
   }

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
      std::cerr << "Sorry, cannot launch the script main routine for " << script_name << std::endl;
      return 1;
   }

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. However, if the VM exited because
   // of runtime errors, the value in A should be ignored.
   if ( ! vm.hadError() )
   {
      Falcon::String str_regA;
      vm.regA().toString( str_regA );

      // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
      Falcon::AutoCString c_regA( str_regA );
      std::cout << "VM Output: " << c_regA.c_str() << std::endl;
   }
   else {
      std::cout << "VM exited with error; we won't show its output. " << std::endl;
   }

   // Get the callable only now, so our script has the chance to change it.
   Falcon::GarbageLock *cbOnUnload = getCallable( &vm, "onUnload" );
   // calling onUnload with the returned item as a parameter.
   if( cbOnUnload != 0 )
   {
      // clear from the previous VM exit.
      vm.resetEvent();

      // push one parameter and call the function.
      vm.pushParameter( vm.regA() );
      vm.callItem( cbOnUnload->item(), 1 ); // one parameter.

      // we're not really interested to the output status of the VM

      // we're not calling cbOnUnload anymore, so this is a good point to
      // allow the VM to dispose it at will
      vm.gcUnlock( cbOnUnload );
   }

   // We are done -- free the resources
   mainMod->decref();

   return 0;
}


Falcon::GarbageLock *getCallable(
      Falcon::VMachine *vm,
      const Falcon::String &callable )
{
   // Find the callable symbol
   const Falcon::Item *call = vm->findGlobalItem( callable );
   if ( call != 0 )
   {
      // if the item is not a callable, we shouldn't bother recording it
      if( ! call->isCallable() )
      {
         Falcon::AutoCString handler( callable );
         std::cerr << "The script provides an " << handler.c_str() <<
            " item, but it's not a callback." << std::endl;
      }
      else {
         // ok, record it.

         // This method will tell the VM that the item is always alive, even if the script
         // destroys it. It returns a GarbageLock that may be used both to access the original
         // item and to free the locked item later on.
         return vm->gcLock( *call );
      }
   }

   // we didn't find a valid callable item
   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );
   
   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   return embed( script_name, argc - 2, argv + 2 );
}

/* end of falcon_embed_cb3.cpp */
