/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_cb4.cpp

   Embedding samples
   VM callback support - script defined callbacks

   This script shows how to enable a script to provide
   register callback functions through application
   provided functions.

   As we're keeping the callback hooks declared by every
   scripts in a list, we need to save some script specific
   data. This is the occasion to show VM overloading and
   script specific data.

   This program will run the main script routine; then it
   will accept strings from user and it will pass them
   to registered command handlers.

   You'll notice that this script does not releases the
   GC Locks acquired to protect the callable items referenced
   outside the VM. That's not an error: they are correctly
   freed at VM destruction. There is no inherent need to
   release a GC Lock, unless it may be safely disposed and
   the VM is still going to run for a long time.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_cb4.cpp -o falcon_embed_cb4

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>
#include <list>
#include <string>


class EmbedVM; // forward decl

// New functions we'll use in this embedding
void processCommandHandlers( EmbedVM &vm );
void processCleanupHandlers( EmbedVM &vm );

//==================================================
// Script specific data
//==================================================

class ScriptData
{
public:
   typedef std::
   list< Falcon::GarbageLock *> t_lockList;

   // this is the list of handlers that will process our commands
   t_lockList m_cmdHandlers;

   // this is the list of handlers that will be called at exit
   t_lockList m_exitHandlers;

   // this is just script-specific data that is used to show
   // how a local function can return some data which is related
   // to the running script
   int m_timesCalled;

   ScriptData():
      m_timesCalled( 0 )
   {}
};


//==================================================
// VM overloading
//==================================================
class EmbedVM: public Falcon::VMachine
{
public:
   ScriptData m_scriptData;
};


//==================================================
// Application module module
//==================================================
// some forward decl

FALCON_FUNC regCallback( Falcon::VMachine *vm );
FALCON_FUNC regCleanup( Falcon::VMachine *vm );
FALCON_FUNC getTimesCalled( Falcon::VMachine *vm );

Falcon::Module *createAppModule()
{
   // create a module and give it an unique name
   Falcon::Module *module = new Falcon::Module;
   module->name( "appModule" );

   // then add the functions
   module->addExtFunc( "regCallback", regCallback );
   module->addExtFunc( "regCleanup", regCleanup );
   module->addExtFunc( "getTimesCalled", getTimesCalled );

   // the module is complete, we can return it
   return module;
}


FALCON_FUNC regCallback( Falcon::VMachine *vm )
{
   // the only parameter is a callable item
   Falcon::Item *iCallable = vm->param(0);
   if( iCallable != 0 && ! iCallable->isCallable() )
   {
      // standard falcon code for parameter declaration: C - Callable
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "C" ) ) );
      return;
   }

   Falcon::GarbageLock *lock = vm->gcLock( *iCallable );

   // We know that we have been called with OUR VM.
   EmbedVM *evm = static_cast<EmbedVM *>( vm );
   ScriptData &sd = evm->m_scriptData;

   // record the handler
   sd.m_cmdHandlers.push_back( lock );
}

FALCON_FUNC regCleanup( Falcon::VMachine *vm )
{
   // the only parameter is a callable item
   Falcon::Item *iCallable = vm->param(0);
   if( iCallable != 0 && ! iCallable->isCallable() )
   {
      // standard falcon code for parameter declaration: C - Callable
      vm->raiseModError( new Falcon::ParamError( Falcon::ErrorParam( Falcon::e_inv_params, __LINE__ ).
            extra( "C" ) ) );
      return;
   }

   Falcon::GarbageLock *lock = vm->gcLock( *iCallable );

   // We know that we have been called with OUR VM.
   EmbedVM *evm = static_cast<EmbedVM *>( vm );
   ScriptData &sd = evm->m_scriptData;

   // record the handler
   sd.m_exitHandlers.push_back( lock );

}

FALCON_FUNC getTimesCalled( Falcon::VMachine *vm )
{
   // We know that we have been called with OUR VM.
   EmbedVM *evm = static_cast<EmbedVM *>( vm );

   // so, all we have to do is to return the desired script-specific data
   vm->retval( (Falcon::int64) evm->m_scriptData.m_timesCalled );

}

//============================================================
// End of the embedded module
//============================================================

// This is the routine that embeds falcon
int embed( const char *script_name, int argc, char **argv )
{

   // Let's create a default error handler
   // It will write on stderr, and the stream will be destroyed with the handler
   Falcon::DefaultErrorHandler err( Falcon::stdErrorStream(), true );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // we take the occasion to show how to load the main module directly,
   // using the module loader:
   Falcon::Module *mainMod = theLoader.loadFile( script_name );
   if ( ! mainMod )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  "." << std::endl;
      return 1;
   }

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );

   // instead of loading it, as we already loaded it, we'll just add it.
   if ( ! rt.addModule( mainMod ) )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  " or its dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   EmbedVM vm;
   // we want the default error handler to intercept also VM errors
   vm.errorHandler( &err );

   vm.link( Falcon::core_module_init() );  // add the core module
   // add also our module:
   Falcon::Module *appModule = createAppModule();
   vm.link( appModule );

   // Just get here a reference to the script data in our VM.
   // A real application may create the script data somewhere else
   // and then store them in the VM.

   // One intelligent thing to do would be to cache the script
   // module ID and save it in the script data, in case it is
   // needed somewhere in the program.
   ScriptData &sd = vm.m_scriptData;


   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   if( ! vm.link( &rt ) )
   {
      std::cerr << "Sorry, cannot link the script and related modules for " << script_name << std::endl;
      return 1;
   }

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm.findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm.findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm.findGlobalItem( "args" );

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::GarbageString( &vm, mainMod->name() );
   *scriptPath = new Falcon::GarbageString( &vm, script_name );

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray( &vm );
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::GarbageString( &vm, argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
      std::cerr << "Sorry, cannot launch the script main routine for " << script_name << std::endl;
      return 1;
   }

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. However, if the VM exited because
   // of runtime errors, the value in A should be ignored.
   if ( ! vm.hadError() )
   {
      Falcon::String str_regA;
      vm.regA().toString( str_regA );

      // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
      Falcon::AutoCString c_regA( str_regA );
      std::cout << "VM Output: " << c_regA.c_str() << std::endl;

      // We can now process user commands, if we have command handlers
      if ( ! sd.m_cmdHandlers.empty() )
      {
         processCommandHandlers( vm );
      }

      processCleanupHandlers( vm );
   }
   else {
      std::cout << "VM exited with error; we won't show its output. " << std::endl;
   }

   // clea
   // We are done -- free the resources
   mainMod->decref();
   appModule->decref();

   return 0;
}


void processCommandHandlers( EmbedVM &vm )
{
   // just a shorthand
   ScriptData &sd = vm.m_scriptData;

   std::cout << "Enter Command: " << std::flush;

   std::string sCmd;
   while( sCmd != "quit" )
   {
      std::getline(std::cin, sCmd);

      if ( sCmd != "" && sCmd != "quit" )
      {
         // for each callback
         ScriptData::t_lockList::const_iterator cbIter = sd.m_cmdHandlers.begin();
         while( cbIter != sd.m_cmdHandlers.end() )
         {
            // passing -1 to GarbageString is a shortcut for bufferization
            // directly in constructor. Not necessary here, as sCmd stays valid,
            // but we do it to show how it works.
            Falcon::GarbageString *command = new Falcon::GarbageString( &vm, sCmd.c_str(), -1 );
            vm.pushParameter( command );
            vm.callItem( (*cbIter)->item(), 1 );

            // ignore even errors
            vm.resetEvent();

            // just to show usage of script data...
            sd.m_timesCalled++;

            ++cbIter;
         }
      }

      if ( sCmd != "quit" )
         std::cout << "Enter Command: " << std::flush;
   }
}


void processCleanupHandlers( EmbedVM &vm )
{
   // just a shorthand
   ScriptData &sd = vm.m_scriptData;

   // Call the cleanup handlers
   ScriptData::t_lockList::const_iterator cbIter = sd.m_exitHandlers.begin();
   while( cbIter != sd.m_exitHandlers.end() )
   {
      vm.callItem( (*cbIter)->item(), 0 );

      // this time, exit in case of error
      if ( vm.hadError() )
         break;

      // just to show usage of script data...
      sd.m_timesCalled++;

      ++cbIter;
   }
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );

   char *script_name;

   if ( argc < 2 ) {
      std::cout << "Please, provide a script name" << std::endl;
      return 0;
   }

   script_name = argv[1];

   // now we also pass the arguments.
   return embed( script_name, argc - 2, argv + 2 );
}

/* end of falcon_embed_cb4.cpp */
