/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_errhand.cpp

   Embedding samples
   Specialized error handler.

   Other than redirecting the default error handler to a stream that
   may be intercepted by the application (see falcon_embed_stream.cpp)
   it is possible to provide a error handler, that will be called
   whenever an error is raised.

   It is then possible to inspect the error and store its representation
   somewhere relevant (i.e. a compile result window, a message box),
   or to alter application status (i.e. set the status of the script
   data representing the running script to "invalid", and update a
   "script list".

   To do fancy things, the error handler should know the vm generating
   the error, or the script that the VM is running. This data is not
   normally provided to the error handler, but the embedder may create
   a class pointing back to the VM where the handler has been set,
   as in the second example we provide.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
               falcon_embed_errhand.cpp -o falcon_embed_errhand

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

//===================================================
// An handler for things happening BEFORE the script
// is loaded and considered executable...
//
class PreErrHandler: public Falcon::ErrorHandler
{
public:
   virtual void handleError( Falcon::Error *error );
};

void PreErrHandler::handleError( Falcon::Error *error )
{
   std::cout << "PreErrHandler::handleError" << std::endl;
   std::cout << "OOOOPPPPSSS!!! --- we had an error while loading the script." << std::endl;
   // let's pick just two relevant data:

   Falcon::AutoCString clsName( error->className() );
   std::cout << "It was an error of class " << clsName.c_str() << ", code " << error->errorCode() << std::endl;

   std::cout << "This is the error as Falcon tell us:" << std::endl;

   Falcon::AutoCString completeError( error->toString() );
   std::cout << completeError.c_str() << std::endl;
}

//===================================================
// and an handler for things happening AFTER the vm
// is running the script...
//
class PostErrHandler: public Falcon::ErrorHandler
{
   // just to show that we may carry script specific data...
   const char *m_script_name;

   // ... or VMs, which in turn may carry script pecific data...
   Falcon::VMachine *m_ownerVM;

public:
   PostErrHandler( Falcon::VMachine *vm, const char *script_name ):
            m_script_name( script_name ),
            m_ownerVM( vm )
   {}

   virtual void handleError( Falcon::Error *error );
};

void PostErrHandler::handleError( Falcon::Error *error )
{
   std::cout << "PostErrHandler::handleError" << std::endl;

   // see: we access specific script data.
   std::cout << "Ops!!! --- Runtime error: " << m_script_name << "." << std::endl;

   // let's pick just two relevant data:

   Falcon::AutoCString clsName( error->className() );
   std::cout << "It was an error of class " << clsName.c_str() << ", code " << error->errorCode() << std::endl;

      // we may access any VM data, including data stored in our
      // own VM subclasses...
   std::cout << "VM is in error state at PC: " << m_ownerVM->programCounter() << std::endl;

   std::cout << "This is the error as Falcon tell us:" << std::endl;
   Falcon::AutoCString completeError( error->toString() );
   std::cout << completeError.c_str() << std::endl;
}

// ===================================================
// End of error handler implementation
//===================================================

// This is the routine that embeds falcon
int embed( const char *script_name )
{
   // let's create the "pre" error handler
   // As this will be shared (or may be shared in a real application),
   // the ownership is left to our program.
   PreErrHandler err;

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler (without ownership)
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // Allow the script to load iteratively other resources it may need.
   Falcon::Runtime rt( &theLoader );
   if ( ! rt.loadFile( script_name ) )
   {
      std::cerr << "Sorry, cannot load the script " << script_name <<  " or its dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachine vm;

   // Now set the error handler for this VM.
   // As this will be specific for this VM, use our own error handler,
   // and tell the VM it is the owner. In this way, it will dispose
   // the handler on destruction.
   vm.errorHandler( new PostErrHandler( &vm, script_name ), true );

   vm.link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   if( ! vm.link( &rt ) )
   {
            std::cerr << "Sorry, cannot link the script and related modules for " << script_name << std::endl;
            return 1;
   }

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
            std::cerr << "Sorry, cannot launch the script main routine for " << script_name << std::endl;
            return 1;
   }

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );
   
   char *script_name;

   if ( argc < 2 ) {
            std::cout << "Please, provide a script name" << std::endl;
            return 0;
   }

   script_name = argv[1];

   return embed( script_name );
}

/* end of falcon_embed_errhand.cpp */
