/*
   FALCON - The Falcon Programming Language.
   FILE: falcon_embed_eval.cpp

   Embedding samples
   Evaluate a string

   This example shows how to use the module loader to compile a string instead of
   an already existing script.

   The first parameter in the command line will be the string to be turned in a script.

   Compile with
      g++ $(falcon-conf --cflags-only-I) $(falcon-conf -L) \
         falcon_embed_eval.cpp -o falcon_embed_eval

   Try to call with something like this:
   ./falcon_embed_eval "for i in [1:10]; > randomPick( args ); end" alpha beta gamma delta

   -------------------------------------------------------------------
   Author: Giancarlo Niccolai
   Begin: 2007-08-11 19:49:00

   -------------------------------------------------------------------
   (C) Copyright 2004: the FALCON developers (see list in AUTHORS file)

   See LICENSE file for licensing details.
*/

// Inclusion of the Falcon Engine
#include <falcon/engine.h>
#include <iostream>

// This is the routine that embeds falcon
int embed( const char *script_data, int argc, char **argv )
{
   // Let's create a default error handler
   // It will write on stderr, and the stream will be destroyed with the handler
   Falcon::DefaultErrorHandler err( Falcon::stdErrorStream(), true );

   // first of all, we need a module loader to load the script.
   // The parameter is the search path for where to search our module
   Falcon::FlcLoader theLoader(".");

   // ... and add the error handler
   theLoader.errorHandler( &err );

   // As we want to use standard Falcon installation,
   // tell the loader that is safe to search module in system path
   theLoader.addFalconPath();

   // now create a string stream that may be used as input:
   Falcon::ROStringStream input( script_data );

   // and use the module loader to create a module on the fly
   Falcon::Module *compiledModule = theLoader.loadSource( &input, "eval" );
   if ( compiledModule == 0 )
   {
      std::cerr << "Sorry, cannot compile the given script." << std::endl;
      return 1;
   }

   // we should give the module a name, as we have created it from thin air
   compiledModule->name( "eval" );

   // Allow the script to load iteratively other resources it may need.
   // A runtime is optional; we don't really need it to compile an
   // on-the-fly script
   Falcon::Runtime rt( &theLoader );

   // now, instead loading the script by path, we just add the compiled module
   if ( ! rt.addModule( compiledModule ) )
   {
      std::cerr << "Sorry, cannot resolve needed dependencies" << std::endl;
      return 1;
   }

   // We are ready to go. Let's create our VM and link in minimal stuff
   Falcon::VMachine vm;
   // we want the default error handler to intercept also VM errors
   vm.errorHandler( &err );

   vm.link( Falcon::core_module_init() );  // add the core module

   // try to link our module and its dependencies.
   // -- It may fail if there are some undefined symbols
   // If you want to remove the runtime, just change &rt with compiledModule
   if( ! vm.link( &rt ) )
   {
      std::cerr << "Sorry, cannot link the script and related modules " << std::endl;
      return 1;
   }

   // Now that we have linked everything, we can set the script name,
   // the script path and the arguments.

   Falcon::Item *scriptName = vm.findGlobalItem( "scriptName" );
   Falcon::Item *scriptPath = vm.findGlobalItem( "scriptPath" );
   Falcon::Item *args = vm.findGlobalItem( "args" );

   // get the topmost (and so, the main) module, just to set the correct name.
   const Falcon::Module *mainMod = vm.mainModule()->module();

   //items can directly be set to Core and Garbage object pointers.
   *scriptName = new Falcon::GarbageString( &vm, mainMod->name() );
   *scriptPath = new Falcon::GarbageString( &vm, "" ); // we have actually no path.

   // create the arguments.
   // It is correct to pass an empty array if we haven't any argument to pass.
   Falcon::CoreArray *argsArray = new Falcon::CoreArray( &vm );
   for( int i = 0; i < argc; i ++ )
   {
      argsArray->append( new Falcon::GarbageString( &vm, argv[i] ) );
   }
   *args = argsArray;

   // end of parameters

   // we're ready to go. Still, we may fail if the script has not a main routine.
   if( ! vm.launch() )
   {
      std::cerr << "Sorry, cannot launch the script main routine " << std::endl;
      return 1;
   }

   // We should have now an output value. It is advisable to turn it
   // in a string before to show it. However, if the VM exited because
   // of runtime errors, the value in A should be ignored.
   if ( ! vm.hadError() )
   {
      Falcon::String str_regA;
      vm.regA().toString( str_regA );

      // Falcon provides a nice helper to convert falcon strings into char * or wchar_t
      Falcon::AutoCString c_regA( str_regA );
      std::cout << "VM Output: " << c_regA.c_str() << std::endl;
   }
   else {
      std::cout << "VM exited with error; we won't show its output. " << std::endl;
   }

   // We are done -- free the resources
   compiledModule->decref();

   return 0;
}


int main( int argc, char *argv[] )
{
   //====================================================
   // Falcon engine initialization.
   Falcon::EngineData data;
   Falcon::Init( data );
   
   // minimal housekeeping
   char *script_data;

   if ( argc < 2 ) {
      std::cout << "Please, provide a whole script in the first parameter." << std::endl;
      return 0;
   }

   script_data = argv[1];

   // now we also pass the arguments.
   return embed( script_data, argc - 2, argv + 2 );
}

/* end of falcon_embed_eval.cpp */
